import json
import os
import platform
from importlib import import_module

from connect.reports.parser import parse
from connect.reports.validator import (
    validate,
    validate_with_schema,
)
from pkg_resources import DistributionNotFound, get_distribution

from executor.exceptions import RunnerException


def get_report(client, report_id):
    return client.ns('reporting').reports[report_id].get()


def get_report_id(func_fqn):
    tokens = func_fqn.split('.')
    if len(tokens) < 3:
        raise Exception("Reports project does not conform with specification")
    return tokens[1]


def get_report_entrypoint(func_fqn):
    module_name, func_name = func_fqn.rsplit('.', 1)
    module = import_module(module_name)
    return getattr(module, func_name)


def get_report_env():
    # Environment variables needed
    report_id = os.getenv('REPORT_ID', None)
    client_token = os.getenv('CLIENT_TOKEN', None)
    api_endpoint = os.getenv('API_ENDPOINT', None)

    required_vars = [
        report_id,
        client_token,
        api_endpoint,
    ]

    if None in required_vars:
        raise Exception("Wrong environment")

    return {
        "report_id": report_id,
        "client_token": client_token,
        "api_endpoint": api_endpoint,
    }


def get_default_reports_dir():
    if os.getenv('REPORTS_MOUNTPOINT'):
        return os.getenv('REPORTS_MOUNTPOINT')

    return os.path.join(
        '/',
        'reports',
        'reports',
    )


def load_descriptor_file(root_path: str):
    descriptor_file = os.path.join(root_path, 'reports.json')
    if not os.path.exists(descriptor_file):
        raise RunnerException('`reports.json` does not exist.')
    try:
        data = json.load(open(descriptor_file, 'r'))
        errors = validate_with_schema(data)
        if errors:
            raise RunnerException(f'Invalid `reports.json`: {errors}')
        repository_definition = parse(root_path, data)
        errors = validate(repository_definition)
        if errors:
            raise RunnerException(f'Invalid `reports.json`: {",".join(errors)}')
        return repository_definition
    except json.JSONDecodeError:
        raise RunnerException('`reports.json` is not a valid json file.')


def get_report_definition(entrypoint):
    root_path = get_default_reports_dir()
    repo_definition = load_descriptor_file(root_path)
    report = [reprt for reprt in repo_definition.reports if reprt.entrypoint == entrypoint][0]
    return report


def get_version():
    try:
        return get_distribution('connect-reports-runner').version
    except DistributionNotFound:
        return '0.0.0'


def get_user_agent():
    version = get_version()
    pimpl = platform.python_implementation()
    pver = platform.python_version()
    sysname = platform.system()
    sysver = platform.release()
    return {
        'User-Agent': (
            f'connect-reports-runner/{version} {pimpl}/{pver} {sysname}/{sysver}'
            f' {os.getenv("REPORT_ID", None)}'
        ),
    }


def upload_file(client, report_name, report_id):
    report_filename = os.path.basename(report_name)
    return client.ns('reporting').reports[report_id].action('upload').post(
        data=open(report_name, 'rb'),
        headers={
            'Content-Type': 'application/octet-stream',
            'Content-Disposition': f'attachment; filename="{report_filename}"',
            **get_user_agent(),
        },
    )



from http import HTTPStatus

from connect.client import ClientError

from executor.utils import get_report, get_report_env


C_SUPPORT = '. Please contact support.'
UNAUTHORIZED = (401, 403)


def format_http_status(status_code):
    status = HTTPStatus(status_code)
    description = status.name.replace('_', ' ').title()
    return f'{status_code} - {description}'


def get_error_message_for_client(res: ClientError):
    status = format_http_status(res.status_code)

    if res.status_code in UNAUTHORIZED:
        return f'{status}: report tried to access objects not accessible by your account'

    if res.status_code == 400:
        code = res.error_code if res.error_code else 'Generic error'
        message = ','.join(res.errors) if res.errors else ''
        return f'{status}: {code} - {message}'
    return f'{status}: unexpected error: {res.message}'


def get_reason(e: Exception, report_type):
    if isinstance(e, ClientError):
        reason = get_error_message_for_client(e)
        if report_type != 'custom':
            reason = reason + C_SUPPORT
    else:
        reason = f'Report execution failed with error: {str(e)}'

    return reason


def report_to_be_blocked(e: Exception, report_type):
    if isinstance(e, RuntimeError):
        return False
    return True if report_type == 'custom' else False


def handle_exception(e: Exception, client, report=None):
    report_env = get_report_env()
    api_endpoint = report_env['api_endpoint']

    if not report:
        report = get_report(client, report_env['report_id'])

    block = report_to_be_blocked(e, report['template']['type'])
    reason = get_reason(e, report['template']['type'])

    if isinstance(e, ClientError):

        exception_cause = e.__cause__

        if exception_cause.request and exception_cause.request.url:  # pragma: no branch
            if e.status_code in UNAUTHORIZED and report['template']['type'] == 'custom':
                block = True
            else:
                # We will not block in the use case of exception caused by using any endpoint
                # More information at LITE-16960
                # to enable back, change the False to True on the else sentence
                block = False if exception_cause.request.url.startswith(api_endpoint) else False

        fail_report(
            client,
            report_env['report_id'],
            reason,
            block,
        )
        raise e
    fail_report(
        client,
        report_env['report_id'],
        reason,
        block,
    )
    raise e


def handle_preparation_exception(e: Exception, client):
    report_env = get_report_env()
    fail_report(
        client,
        report_env['report_id'],
        "An error happened while preparing report execution, please try again later or contact "
        "support",
        False,
    )

    raise e


def handle_post_execution_exception(e: Exception, client):
    report_env = get_report_env()
    report_id = report_env['report_id']
    fail_report(
        client,
        report_id,
        f'Error storing the report{C_SUPPORT}',
        False,
    )
    raise e


def fail_report(client, report_id, reason, block):
    return client.ns('reporting').reports[report_id].action('fail').post(
        {
            "notes": reason,
            "block": block,
        },
    )



import logging
import sys
from datetime import datetime

import pytz
from connect.client import ClientError, ConnectClient
from connect.reports.constants import REPORTS_ENV
from connect.reports.datamodels import Account, Report
from connect.reports.renderers import get_renderer

from executor.exception_handler import (
    handle_exception,
    handle_post_execution_exception,
    handle_preparation_exception,
)
from executor.utils import (
    get_default_reports_dir,
    get_report,
    get_report_definition,
    get_report_entrypoint,
    get_report_env,
    get_user_agent,
    upload_file,
)


logger = logging.getLogger('executor')


def start():
    logger.info("Preparing environment for report execution")
    report_env = get_report_env()

    client = ConnectClient(
        endpoint=report_env['api_endpoint'],
        use_specs=False,
        api_key=report_env["client_token"],
        max_retries=5,
        default_limit=500,
        default_headers=get_user_agent(),
        timeout=(360, 360),
    )

    try:
        report_to_execute = get_report(client, report_env["report_id"])
        logger.info(f"Preparing execution of report {report_to_execute}")
        report_definition = get_report_definition(report_to_execute['template']['entrypoint'])

    except (ClientError, Exception) as e:
        logger.exception('An error ocurred while preparing the execution environment.')
        handle_preparation_exception(e, client)

    result = execute_report(
        control_client=client,
        report_definition=report_definition,
        connect_report=report_to_execute,
    )

    if result:  # pragma: no branch
        try:
            upload_file(client, result, report_env["report_id"])
        except (ClientError, Exception) as e:
            logger.exception('An error ocurred during report upload.')
            handle_post_execution_exception(e, client)


def normalize_parameters(connect_parameters):
    parameters = {}
    for param in connect_parameters:
        parameters[param['id']] = param['value']

    return parameters


def execute_report(control_client, report_definition, connect_report):  # noqa: CCR001
    report_env = get_report_env()
    reports_dir = get_default_reports_dir()

    report_client = ConnectClient(
        endpoint=report_env["api_endpoint"],
        use_specs=False,
        api_key=report_env["client_token"],
        max_retries=5,
        default_limit=250,
        default_headers=get_user_agent(),
        timeout=(180, 1500),
        resourceset_append=False,
    )
    connect_parameters = connect_report.get('parameters', [])
    parameters = normalize_parameters(connect_parameters)

    def progress(current_value, max_value):
        report = False
        if (
            max_value < 100 and current_value % 10 == 0
            or max_value < 1000 and current_value % 20 == 0
            or max_value < 10000 and current_value % 50 == 0
            or current_value % 2000 == 0
        ):
            report = True
        if report:  # pragma: no branch
            control_client.ns(
                'reporting',
            ).reports[report_env['report_id']].action(
                'progress',
            ).post(
                {
                    "progress": {
                        "max": max_value,
                        "value": current_value,
                    },
                },
            )

    if reports_dir not in sys.path:
        sys.path.append(reports_dir)
    try:
        report_entry_point = get_report_entrypoint(report_definition.entrypoint)
    except (ImportError, AttributeError) as e:
        logger.exception('An error ocurred while importing report entrypoint.')
        handle_preparation_exception(e, control_client)

    renderer_id = connect_report['renderer']
    renderer_definition = next(
        filter(
            lambda renderer: renderer.id == renderer_id, report_definition.renderers,
        ),
    )

    renderer = get_renderer(
        renderer_definition.type,
        REPORTS_ENV,
        reports_dir,
        Account(connect_report['owner']['id'], connect_report['owner']['name']),
        Report(
            report_definition.local_id,
            report_definition.name,
            report_definition.description,
            parameters,
        ),
        renderer_definition.template,
        renderer_definition.args,
    )

    try:
        args = [report_client, parameters, progress]
        if report_definition.report_spec == '2':
            args.extend(
                [
                    renderer_definition.type,
                    renderer.set_extra_context,
                ],
            )
        data = report_entry_point(*args)
        return renderer.render(data, '/report', start_time=datetime.now(tz=pytz.utc))
    except Exception as e:
        handle_exception(e, control_client, connect_report)


# Launch main process
if __name__ == '__main__':  # pragma: no cover
    logging.basicConfig(
        level=logging.INFO,
        format='%(asctime)s %(name)s %(levelname)s PID_%(process)d %(message)s',
    )
    try:  # pragma: no cover
        start()
    except BaseException:
        logger.critical('Unhandled exception has ocurred.', exc_info=True)



import logging
import subprocess

from connect.client import ClientError, ConnectClient

from executor.exception_handler import fail_report
from executor.utils import get_report_env, get_user_agent


logger = logging.getLogger('runner')


def run_executor():
    logging.basicConfig(
        level=logging.INFO,
        format='%(asctime)s %(name)s %(levelname)s PID_%(process)d %(message)s',
    )

    report_env = get_report_env()
    report_id = report_env['report_id']

    proc = subprocess.Popen(
        [
            'python',
            '-m',
            'executor.executor',
        ],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    logger.info(f'Executor started: PID {proc.pid}, report: {report_id}')
    stdout, stderr = proc.communicate()

    if proc.returncode == 0:
        logger.info('Executor process has exited with 0.')
        return

    logger.error(f'Executor process has exited with {proc.returncode}: {stdout=} {stderr=}.')

    client = ConnectClient(
        endpoint=report_env['api_endpoint'],
        use_specs=False,
        api_key=report_env['client_token'],
        max_retries=3,
        default_headers=get_user_agent(),
        default_limit=500,
    )
    try:
        fail_report(
            client,
            report_id,
            'Report execution has failed due contains too much data, please try to exclude '
            'using report parameters some of it and try again.',
            False,
        )
        logger.info(f'Report {report_id} has been failed successfully.')
    except ClientError as ce:
        logger.warning(f'Cannot switch report {report_id} to fail status: {ce}')
    return


if __name__ == '__main__':  # pragma: no cover
    run_executor()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['executor']

package_data = \
{'': ['*']}

install_requires = \
['chardet>=3.0.0,<4.0.0',
 'connect-openapi-client>=25.4',
 'connect-reports-core>=25.0.0,<26.0.0',
 'lxml>=4.0.0,<5.0.0',
 'openpyxl>=3.0.0,<4.0.0',
 'requests>=2.0.0,<3.0.0']

entry_points = \
{'console_scripts': ['cextrun = executor.runner:run_executor']}

setup_kwargs = {
    'name': 'connect-reports-runner',
    'version': '25.6',
    'description': 'Connect Reports Runner',
    'long_description': 'Connect Reports Executor\n========================\n\nThis project contains the executor code used to run custom reports on CloudBlue Connect.\n',
    'author': 'CloudBlue LLC',
    'author_email': 'None',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://connect.cloudblue.com',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
